# (c) 2009-2024 Martin Wendt and contributors; see WsgiDAV https://github.com/mar10/wsgidav
# Licensed under the MIT license: http://www.opensource.org/licenses/mit-license.php
"""
Implementation of a domain controller that allows users to authenticate against
a Pluggable Authentication Module ('PAM').

Used by HTTPAuthenticator. Only available on linux and macOS.

See https://wsgidav.readthedocs.io/en/latest/user_guide_configure.html
"""

import os
import threading

from wsgidav import util
from wsgidav.dc.base_dc import BaseDomainController

__docformat__ = "reStructuredText"
_logger = util.get_module_logger(__name__)

try:
    import pam
except ImportError:
    _logger.error(
        "pam_dc requires the `python-pam` module. Try `pip install wsgidav[pam]`."
    )
    raise


class PAMDomainController(BaseDomainController):
    def __init__(self, wsgidav_app, config):
        super().__init__(wsgidav_app, config)

        self.lock = threading.RLock()
        self.pam = pam.pam()

        dc_conf = util.get_dict_value(config, "pam_dc", as_dict=True)

        self.pam_service = dc_conf.get("service", "login")
        self.pam_encoding = dc_conf.get("encoding", "utf-8")
        self.pam_resetcreds = dc_conf.get("resetcreds", True)
        self.allow_users = dc_conf.get("allow_users", "all")
        if not (
            self.allow_users in ("all", "current") or isinstance(self.allow_users, list)
        ):
            raise ValueError(
                f"Invalid 'allow_users' value: {self.allow_users!r}, expected 'all', 'current' or list of allowed users."
            )
        self.deny_users = dc_conf.get("deny_users", [])
        if not isinstance(self.deny_users, list):
            raise ValueError(
                f"Invalid 'deny_users' value: {self.deny_users!r}, expected list of denied users."
            )

    def __str__(self):
        return f"{self.__class__.__name__}({self.pam_service!r})"

    def get_domain_realm(self, path_info, environ):
        return f"PAM({self.pam_service})"

    def require_authentication(self, realm, environ):
        return True

    def _validate_user(self, user_name):
        if user_name in self.deny_users:
            return False
        if self.allow_users == "all":
            return True
        if self.allow_users == "current":
            if user_name == os.getlogin():
                return True
        if user_name in self.allow_users:
            return True
        return False

    def basic_auth_user(self, realm, user_name, password, environ):
        # Seems that python_pam is not threadsafe (#265)
        if not self._validate_user(user_name):
            _logger.warning(f"User {user_name!r} is not allowed.")
            return False
        with self.lock:
            is_ok = self.pam.authenticate(
                user_name,
                password,
                service=self.pam_service,
                resetcreds=self.pam_resetcreds,
                encoding=self.pam_encoding,
            )
            if not is_ok:
                _logger.warning(
                    f"pam.authenticate({user_name!r}, '<redacted>', {self.pam_service!r}) failed with code {self.pam.code}: {self.pam.reason}"
                )
                return False

        _logger.debug(f"User {user_name!r} logged on.")
        return True

    def supports_http_digest_auth(self):
        # We don't have access to a plaintext password (or stored hash)
        return False



from passlib.apache import HtpasswdFile

from wsgidav import util
from wsgidav.dc.base_dc import BaseDomainController


class HtpasswdDomainController(BaseDomainController):
    def __init__(self, wsgidav_app, config):
        super().__init__(wsgidav_app, config)

        dc_conf = util.get_dict_value(config, "htpasswd_dc", as_dict=True)
        file_path = dc_conf.get("htpasswd_file")

        if file_path is None:
            raise RuntimeError("Missing option: htpasswd_dc.htpasswd_file")

        self.htpasswd = HtpasswdFile(file_path)

    def __str__(self):
        return f"{self.__class__.__name__}()"

    def get_domain_realm(self, path_info, environ):
        realm = self._calc_realm_from_path_provider(path_info, environ)
        return realm

    def require_authentication(self, realm, environ):
        return True

    def supports_http_digest_auth(self):
        return False

    def basic_auth_user(self, realm, user_name, password, environ):
        return self.htpasswd.check_password(user_name, password)



"""
Current WsgiDAV version number.

See https://www.python.org/dev/peps/pep-0440

Examples
    Pre-releases (alpha, beta, release candidate):
        '3.0.0a1', '3.0.0b1', '3.0.0rc1'
    Final Release:
        '3.0.0'
    Developmental release (to mark 3.0.0 as 'used'. Don't publish this):
        '3.0.0.dev1'
NOTE:
    When pywin32 is installed, number must be a.b.c for MSI builds?
    "3.0.0a4" seems not to work in this case!
"""

__version__ = "4.3.5"

# make version accessible as 'wsgidav.__version__'
# from wsgidav._version import __version__  # noqa: F401

# Initialize a silent 'wsgidav' logger
# http://docs.python-guide.org/en/latest/writing/logging/#logging-in-a-library
# https://docs.python.org/3/howto/logging.html#configuring-logging-for-a-library
import logging

_base_logger = logging.getLogger(__name__)
_base_logger.addHandler(logging.NullHandler())
_base_logger.propagate = False
_base_logger.setLevel(logging.INFO)



# (c) 2009-2024 Martin Wendt and contributors; see WsgiDAV https://github.com/mar10/wsgidav
# Licensed under the MIT license:
# http://www.opensource.org/licenses/mit-license.php
"""
WSGI middleware to impersonate a Unix user based on the HTTP username.
This middleware changes the effective user ID (euid) and group ID (egid) of the
current process to match the Unix user corresponding to the HTTP username.
This is useful for applications that need to access files or resources with the
permissions of a specific Unix user.

NOTE: **Experimental**:
This middleware is not thread-safe and should not be used in a
multithreaded environment. It is designed to be used in a single-threaded
context where the effective user ID and group ID can be safely changed and
restored.
"""

import os
import pwd
from contextlib import AbstractContextManager
from typing import Tuple

from wsgidav import util
from wsgidav.mw.base_mw import BaseMiddleware

_logger = util.get_module_logger(__name__)
_init_euid = os.geteuid()
_init_egid = os.getegid()
_logger.debug(f"impersonator: initial uid:gid = {_init_euid}:{_init_egid}")


class ImpersonateContext(AbstractContextManager):
    def __init__(self, ids: Tuple[int, int] | None) -> None:
        if ids is None:
            self._enabled = False
            return
        self._enabled = True
        self._new_euid = ids[0]
        self._new_egid = ids[1]
        self._old_euid = os.geteuid()
        self._old_egid = os.getegid()

        if self._old_euid != _init_euid or self._old_egid != _init_egid:
            raise Exception(
                "old ids mismatched with init ids: "
                f"{self._old_euid}:{self._old_egid} versus {_init_euid}:{_init_egid}, "
                "multithreading MUST be disabled for impersonator to function correctly"
            )

    def __enter__(self):
        if not self._enabled:
            return
        os.seteuid(self._new_euid)
        os.setegid(self._new_egid)
        _logger.debug(f"impersonator: set uid:gid = {self._new_euid}:{self._new_egid}")

    def __exit__(self, exc_type, exc_value, traceback, /):
        if not self._enabled:
            return
        os.seteuid(self._old_euid)
        os.setegid(self._old_egid)
        _logger.debug(
            f"impersonator: reset uid:gid = {self._old_euid}:{self._old_egid}"
        )


class Impersonator(BaseMiddleware):
    def __init__(self, wsgidav_app, next_app, config):
        super().__init__(wsgidav_app, next_app, config)

    def __call__(self, environ, start_response):
        username = environ.get("wsgidav.auth.user_name", "")
        ids = self._map_id(username)
        with ImpersonateContext(ids):
            yield from self.next_app(environ, start_response)

    def is_disabled(self):  # type: ignore
        return not self.get_config("impersonator.enable", False)  # type: ignore

    def _map_id(self, username: str) -> Tuple[int, int] | None:
        if self.is_disabled():
            return None

        unix_username = None

        if not username:
            unix_username = "nobody"
        elif self.get_config("impersonator.custom_user_mapping", None) is None:  # type: ignore
            unix_username = username
        else:
            unix_username = self.get_config("impersonator.custom_user_mapping").get(
                username, None
            )  # type: ignore

        if unix_username is None:
            raise RuntimeError(
                f"Failed mapping HTTP username '{username}' to Unix username"
            )

        _logger.debug(
            f"impersonator: HTTP user {username or '(anonymous)'} -> Unix user {unix_username}"
        )

        try:
            passwd = pwd.getpwnam(unix_username)
        except Exception:
            raise RuntimeError(
                f"Unix username '{unix_username}' does not exist"
            ) from None
        else:
            _logger.debug(
                f"impersonator: Unix user {unix_username} -> uid:gid = {passwd.pw_uid}:{passwd.pw_gid}"
            )
            return (passwd.pw_uid, passwd.pw_gid)
